#
# Copyright (c) 2023 Commonwealth Scientific and Industrial Research Organisation (CSIRO). All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file. See the AUTHORS file for names of contributors.
#
import asyncio
import concurrent.futures
import traceback
from typing import Any, Callable, Generic, Optional, TypeVar, Union
from cachetools import TTLCache
from fastapi import Request
from pydantic import BaseModel, Field
from ivcap_fastapi import getLogger
from opentelemetry import trace

from .utils import _get_input_type

logger = getLogger("executor")
tracer = trace.get_tracer("executor")

class ExecutionContext(BaseModel):
    pass

T = TypeVar('T')

class ExecutorOpts(BaseModel):
    job_cache_size: Optional[int] = Field(1000, description="size of job cache")
    job_cache_ttl: Optional[int] = Field(600, description="TTL of job entries in the job cache")
    max_workers: Optional[int] = Field(None, description="size of thread pool to use. If None, a new thread pool will be created for each execution")

class ExecutionError(BaseModel):
    """
    Pydantic model for execution errors.
    """
    error: str
    type: type # type of error
    traceback: str

class Executor(Generic[T]):
    """
    A generic class that executes a function in a thread pool and returns the result via an asyncio Queue.
    The generic type T represents the return type of the function.
    """

    def __init__(
        self,
        func: Callable[..., T],
        *,
        opts: Optional[ExecutorOpts],
        context: Optional[ExecutionContext] = None
    ):
        """
        Initialize the Executor with a function and an optional thread pool.

        Args:
            func: The function to execute, returning type T
            opts:
             - job_cache_size: Optional size of job cache. Defaults to 1000
             - job_cache_ttl: Optional TTL of job entries in the job cache. Defaults to 600 sec
             - max_workers: Optional size of thread pool to use. If None, a new thread pool will be created for each execution.
        """
        self.func = func
        if opts is None:
            opts = ExecutorOpts()
        self.job_cache = TTLCache(maxsize=opts.job_cache_size, ttl=opts.job_cache_ttl)
        self.thread_pool = None
        if opts.max_workers:
            self.thread_pool = concurrent.futures.ThreadPoolExecutor(max_workers=opts.max_workers)

        self.context = context
        self.context_param = None
        self.request_param = None
        _, extras = _get_input_type(func)
        for k, v in extras.items():
            if isinstance(context, v):
                self.context_param = k
            if v == Request:
                self.request_param  = k

    async def execute(self, param: Any, job_id: str, req: Request) -> asyncio.Queue[Union[T, ExecutionError]]:
        """
        Execute the function with the given parameter in a thread and return a queue with the result.

        Args:
            param: Any The parameter to pass to the function
            job_id: str ID of this job
            req: Request FastAPI's request object

        Returns:
            An asyncio Queue that will contain either the result of type T or an ExecutionError
        """
        result_queue: asyncio.Queue[Union[T, ExecutionError]] = asyncio.Queue()
        event_loop = asyncio.get_event_loop()
        self.job_cache[job_id] = None
        def _callback(future):
            """Callback to handle the future result and put it in the queue."""
            try:
                result = future.result()
            except Exception as e:
                result = ExecutionError(
                    error=str(e),
                    type=type(e),
                    traceback=traceback.format_exc()
                )
                logger.warning(f"job {job_id} failed - {result.error}")
            finally:
                self.job_cache[job_id] = result
                asyncio.run_coroutine_threadsafe(
                    result_queue.put(result),
                    event_loop,
                )

        def _run(param: Any):
            kwargs = {}
            if self.context_param is not None:
                kwargs[self.context_param] = self.context
            if self.request_param is not None:
                kwargs[self.request_param] = req

            fname = self.func.__name__
            with tracer.start_as_current_span(f"RUN {fname}") as span:
                span.set_attribute("job.id", job_id)
                span.set_attribute("job.name", fname)
                try:
                    if asyncio.iscoroutinefunction(self.func):
                        loop = asyncio.new_event_loop()
                        asyncio.set_event_loop(loop)
                        res = loop.run_until_complete(self.func(param, **kwargs))
                        loop.close()  # Clean up event loop
                        return res
                    else:
                        return self.func(param, **kwargs)
                except Exception as ex:
                    span.record_exception(ex)
                    raise ex


        # Use the provided thread pool or create a new one
        use_pool = self.thread_pool or concurrent.futures.ThreadPoolExecutor(max_workers=1)

        # Submit the function to the thread pool
        future = use_pool.submit(_run, param)
        future.add_done_callback(_callback)

        # If we created a new pool, we should clean it up when done
        if self.thread_pool is None:
            future.add_done_callback(lambda _: use_pool.shutdown(wait=False))

        return result_queue

    def lookup_job(self, job_id: str) -> Union[T, ExecutionError, None]:
        """Return the result of a

        Args:
            job_id (str): The id of the job requested

        Returns:
            Union[T, ExecutionError, None]: Returns the result fo a job, 'None' is still in progress

        Raises:
            KeyError: Unknown job - may have already expired
        """
        return self.job_cache[job_id]



#
# Copyright (c) 2023 Commonwealth Scientific and Industrial Research Organisation (CSIRO). All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file. See the AUTHORS file for names of contributors.
#
import inspect
import re
from typing import Optional, Type, Callable, TypeVar, Any, get_type_hints, Union, Dict, Tuple

from pydantic import BaseModel


def _get_title_from_path(path: str) -> Tuple[str, str]:
    """Extracts a title from a path string.

    Args:
        path: A string potentially formatted as a directory path (separated by '/').

    Returns:
        A tuple containing:
        - The processed string with the first character in lowercase.
        - The processed string with the first character in uppercase.
    """
    # Extract the last element of the path
    last_element = path.split('/')[-1]

    # Convert to singular form if plural
    singular = last_element
    if singular.endswith('ies'):
        singular = singular[:-3] + 'y'
    elif singular.endswith('es'):
        singular = singular[:-2]
    elif singular.endswith('ss'):
        pass
    elif singular.endswith('s'):
        singular = singular[:-1]

    # Replace underscores with spaces
    singular = singular.replace('_', ' ')

    # Create lowercase and uppercase versions
    if singular:
        lowercase = singular[0].lower() + singular[1:] if len(singular) > 1 else singular.lower()
        uppercase = singular[0].upper() + singular[1:] if len(singular) > 1 else singular.upper()
    else:
        lowercase = ""
        uppercase = ""

    return (lowercase, uppercase)


def _get_input_type(func: Callable) -> Tuple[Optional[Type[BaseModel]], Dict[str, Any]]:
    """Gets the input type of a function.

    Args:
        func: The function to get the input type for.

    Returns:
        A tuple containing:
        - The first function parameter which is a derived class of a pydantic BaseModel, or None if no such parameter exists.
        - A dictionary of all additional parameters, where the key is the parameter name and the value is the type.
    """
    signature = inspect.signature(func)
    type_hints = get_type_hints(func)

    # Get the Pydantic model class
    pydantic_model_class = None
    pydantic_param_name = None
    for param_name, param in signature.parameters.items():
        if hasattr(param.annotation, '__mro__') and BaseModel in param.annotation.__mro__:
            pydantic_model_class = param.annotation
            pydantic_param_name = param_name
            break

    # Get all additional parameters
    additional_params = {}
    for param_name, param in signature.parameters.items():
        if param_name != pydantic_param_name:
            param_type = type_hints.get(param_name, Any)
            additional_params[param_name] = param_type

    return pydantic_model_class, additional_params

def _get_function_return_type(func):
    """Extracts the return type from a function."""
    type_hints = get_type_hints(func)
    # param_types = {k: v for k, v in type_hints.items() if k != 'return'}
    return_type = type_hints.get('return')
    # return param_types, return_type
    return return_type

def find_first(iterable, condition):
    """
    Returns the first item in the iterable for which the condition is True.
    If no such item is found, returns None.
    """
    for item in iterable:
        if condition(item):
            return item
    return None



#
# Copyright (c) 2023 Commonwealth Scientific and Industrial Research Organisation (CSIRO). All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file. See the AUTHORS file for names of contributors.
#
import asyncio
from fastapi import FastAPI, Response, status, Request
from pydantic import BaseModel, Field
from typing import Any, Dict, Optional, Callable, TypeVar
from uuid6 import uuid6

from ivcap_fastapi import getLogger, TryLaterException

from .executor import ExecutionContext, ExecutionError, Executor, ExecutorOpts
from .utils import _get_input_type, _get_function_return_type, _get_title_from_path
from .tool_definition import ToolDefinition, create_tool_definition



class ErrorModel(BaseModel):
    message: str
    code: int

class ExecutionErrorModel(BaseModel):
    jschema: str = Field("urn:ivcap:schema.ai-tool.error.1", alias="$schema")
    message: str
    traceback: str


logger = getLogger("wrapper")

class ToolOptions(BaseModel):
    name: Optional[str] = Field(None, description="Name to be used for this tool")
    tags: Optional[list[str]] = Field(None, description="OpenAPI tag for this set of functions")
    max_wait_time: Optional[float] = Field(5.0, description="max. time in seconds to wait for result and before returning RetryLater")
    refresh_interval: Optional[int] = Field(3, description="Time in seconds to wait before chacking again for a job result (used in RetryLater)")
    executor_opts: Optional[ExecutorOpts] = Field(None, description="Options for the executor")
    post_route_opts: Optional[Dict[str, Any]] = Field({}, description="Addtitional options given the POST route constructor")

# Define a generic type for Pydantic models
T = TypeVar("T", bound=BaseModel)

def add_tool_api_route(
    app: FastAPI,
    path_prefix: str,
    worker_fn: Callable[[BaseModel, Optional[ExecutionContext], Optional[Response]], BaseModel],
    *,
    opts: Optional[ToolOptions] = ToolOptions(),
    context: Optional[ExecutionContext] = None
):
    """Add a few routes to `app` for use with an AI tool.

    The tool itself is implemented in `worker_fn` where the first
    argument is a pydantic model describing all the tool's "public" parameters.
    The function is also expected to return it's result as a single pydantic model.
    The tool function can have two optioanl paramters, one with the same type as
    `context` and the second one with `fastapi.Request`. The context paramter will
    be identical to the above `context`, while the `request` will be the incoming
    request.

    This function then sets up three endpoints:
    - POST {path_prefix}: To request the execution of the tool (the 'job')
    - GET {path_prefix}/{job_id}: To collect the result of a tool execution
    - GET {path_prefix}: To obtain a description of the tool suitable for most agent frameworks

    The POST request will only wait `opts.max_wait_time` for the tool to finish. If it
    hasn't finished by then, a `204 No Content` code will be returned with additional
    header fields `Location` and `Retry-later` to inform the caller where and approx.
    when the result can be collected later.

    The `opts` parameter allows for customization of the endpoints. See `ToolOptions`
    for a more detailed description.

    Args:
        app (FastAPI): The FastAPI context
        path_prefix (str): The path prefix to use for this set of endpoints
        worker_fn (Callable[[BaseModel, Optional[ExecutionContext], Optional[Response]], BaseModel]): _description_
        opts (Optional[ToolOptions], optional): Additional behaviour settings. Defaults to ToolOptions().
        context (Optional[ExecutionContext], optional): An optional context to be provided to every invocation of `worker_fn`. Defaults to None.
    """
    def_name, def_tag = _get_title_from_path(path_prefix)
    if opts.tags is None:
        if def_tag == '':
            def_tag = "Tool"
        opts.tags = [def_tag]
    if opts.name is None:
        if def_name == '':
            def_name = "Execute the tool"
        opts.name = def_name

    output_model = _get_function_return_type(worker_fn)
    executor = Executor[output_model](worker_fn, opts=opts.executor_opts, context=context)

    _add_do_job_route(app, path_prefix, worker_fn, executor, opts)
    _add_get_job_route(app, path_prefix, worker_fn, executor, opts)
    _add_get_tool_def_route(app, path_prefix, worker_fn, opts)

def _add_do_job_route(app: FastAPI, path_prefix: str, worker_fn: Callable, executor: Executor, opts: ToolOptions):
    input_model, _ = _get_input_type(worker_fn)
    output_model = _get_function_return_type(worker_fn)
    summary, description = (worker_fn.__doc__.lstrip() + "\n").split("\n", 1)

    async def route(data: input_model, req: Request) -> output_model:  # type: ignore
        job_id = req.headers.get("job-id")
        if job_id == None:
            job_id = str(uuid6())

        if req.headers.get("prefer") == "respond-async":
            timeout = 0
        else:
            toh = req.headers.get("timeout")
            if toh != None:
                timeout = int(toh)
            else:
                timeout = opts.max_wait_time

        queue = await executor.execute(data, job_id, req)
        try:
            el = await asyncio.wait_for(queue.get(), timeout=timeout)
            queue.task_done()
            el = _return_job_result(el, job_id)
            return el
        except asyncio.TimeoutError:
            return _return_try_later(job_id, path_prefix, opts)

    responses = {
        204: {
            "headers": {
                "Location": {
                    "description": "The URL where to pick up the result of this request",
                    "type": "string",
                },
                "Retry-Later": {
                    "description": "The time to wait before checking for a result",
                    "type": "integer",
                },
            },
        },
        400: { "model": ErrorModel, },
        # 400: {"model": Error}, 401: {"model": Error}, 429: {"model": Error}},
    }
    app.add_api_route(
        path_prefix,
        route,
        # name=opts.name,
        summary=summary,
        description=description.strip(),
        methods=["POST"],
        responses=responses,
        tags=opts.tags,
        response_model_exclude_none=True,
        response_model_by_alias=True,
        **opts.post_route_opts,
    )

def _add_get_job_route(app: FastAPI, path_prefix: str, worker_fn: Callable, executor: Executor, opts: ToolOptions):
    output_model = _get_function_return_type(worker_fn)
    def route(job_id: str) -> output_model: # type: ignore
        try:
            result = executor.lookup_job(job_id)
            if result == None:
                return _return_try_later(job_id, path_prefix, opts)
            return _return_job_result(result, job_id)
        except KeyError:
            return Response(status_code=status.HTTP_404_NOT_FOUND,
                            content=f"job {job_id} can't be found. It either never existed or its result is no longer cached.")

    responses = {
        400: { "model": ErrorModel, },
    }
    path = "/jobs/" + "{job_id}"
    if path_prefix != "/":
        path = path_prefix + path
    app.add_api_route(
        path,
        route,
        summary="Returns the result of a particular job.",
        methods=["GET"],
        responses=responses,
        tags=opts.tags,
        response_model_exclude_none=True,
        response_model_by_alias=True,
    )

def _return_job_result(el, job_id):
    if isinstance(el, ExecutionError):
        if el.type == ValueError:
            m = ErrorModel(message=el.error, code=400)
            return Response(status_code=status.HTTP_400_BAD_REQUEST, content=m.model_dump_json(indent=2), media_type="application/json")

        m = ExecutionErrorModel(message=el.error, traceback=el.traceback)
        return  Response(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, content=m.model_dump_json(indent=2), media_type="application/json")
    else:
        return el

def _add_get_tool_def_route(app: FastAPI, path_prefix: str, worker_fn: Callable, opts: ToolOptions):
    async def route() -> ToolDefinition:  # type: ignore
        return create_tool_definition(worker_fn, name=opts.name)

    app.add_api_route(
        path_prefix,
        route,
        summary="Returns the description of this tool. Primarily used by agents.",
        methods=["GET"],
        tags=opts.tags,
        response_model_exclude_none=True,
        response_model_by_alias=True,
    )


def _return_try_later(job_id: str, path_prefix: str, opts: ToolOptions):
    location = f"/jobs/{job_id}"
    if path_prefix != "/":
        location = path_prefix + location
    headers = {
        "Location": location,
        "Retry-Later": f"{opts.refresh_interval}",
    }
    return Response(status_code=status.HTTP_204_NO_CONTENT, headers=headers)



#
# Copyright (c) 2023 Commonwealth Scientific and Industrial Research Organisation (CSIRO). All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file. See the AUTHORS file for names of contributors.
#
""" A client library for accessing IVCAP """

# read version from installed package
try:  # Python < 3.10 (backport)
    from importlib_metadata import version
except ImportError:
    from importlib.metadata import version
try:
    __version__ = version("ivcap_client")
except Exception:
    __version__ = "unknown" # should only happen when running the local examples

from .tool_definition import ToolDefinition, print_tool_definition, create_tool_definition
from .server import start_tool_server

from .builder import add_tool_api_route, ToolOptions
from .executor import ExecutionContext